/**
 *  Time-stamp:  <2011-01-26 18:35:16 raskolnikov>
 *
 *  @file        bnb.tpp
 *  @author      Juan Pedro Bolivar Puente <raskolnikov@es.gnu.org>
 *  @date        Thu Jan 13 18:48:44 2011
 *
 *  Implementacion del branch and bound.
 */

#ifndef PDP_BNB_TPP_
#define PDP_BNB_TPP_

namespace pdp
{
namespace bnb
{

template <typename Node>
void state<Node>::bound ()
{
    typename state<Node>::stack_type::iterator it = _stack.begin ();
    const cost_type threshold = best_cost ();
    
    while (it != _stack.end ())
    {
        if (it->cost () >= threshold)
            it = _stack.erase (it);
        else
            ++it;
    }
}

template <typename Node>
void state<Node>::push (const Node& n)
{
    if (bnb_is_solution (n))
    {
        if (bnb_cost (n) < best_cost ())
            _best_node = n;
    }
    else if (bnb_cost (n) < best_cost ())
    {
#if 0
        std::cout << n;
#endif
        _stack.push_front (n);
    }
}

template <typename Node>
void state<Node>::push_back (const Node& n)
{
    if (bnb_is_solution (n))
    {
        if (bnb_cost (n) < best_cost ())
            _best_node = n;
    }
    else if (bnb_cost (n) < best_cost ())
    {
#if 0
        std::cout << n;
#endif
        _stack.push_back (n);
    }
}

template <typename Node>
void state<Node>::pop ()
{
    _next = _stack.front ();
    _stack.pop_front ();
}

template <typename Node, typename Context>
Node branch_and_bound (Context& ctx)
{
    typedef typename node_traits<Node>::cost_type cost_type;
    state<Node> stack (node_traits<Node>::initial (ctx));
    
    while (!stack.empty ())
    {
        stack.pop ();
        
        cost_type old_cost = stack.best_cost ();
        stack.next ().branch (ctx, stack);
        
        if (old_cost != stack.best_cost ())
            stack.bound ();
    }

    return stack.best_node ();
}

} /* namespace bnb */
} /* namespace pdp */

#endif /* PDP_BNB_TPP_ */

