/**
 *  Time-stamp:  <2010-11-25 19:17:14 raskolnikov>
 *
 *  @file        uni_dist_graph.tpp
 *  @author      Juan Pedro Bolivar Puente <raskolnikov@es.gnu.org>
 *  @date        Fri Nov 19 11:43:06 2010
 *
 *  Implementación.
 */

#ifndef PDP_UNI_DIST_GRAPH_TPP_
#define PDP_UNI_DIST_GRAPH_TPP_

#include "full_dist_graph.hpp"

namespace pdp
{

template <typename Label>
void uni_dist_graph<Label>::_check (MPI::Comm& comm, graph<Label>& g)
{
    if (comm.Get_size () > (int) g.size () ||
	g.size () % comm.Get_size () != 0)
	throw std::runtime_error ("Numero de proc. incorrecto.");
}

template <typename Label>
void uni_dist_graph<Label>::scatter (MPI::Comm& comm, graph<Label>& g, int root)
{
    detail::bcast_graph_size (comm, g, root);
    _check (comm, g);
    
    // Guardamos los datos necesiarios para realizar la distribución.
    _comm         = &comm;
    _default_root = root;
    _data         = g.raw ();
    _size         = g.size ();
    _local_range (comm, _local_begin, _local_end);
    size_t rows = _local_end - _local_begin;

    // Distribuimos.
    comm.Scatter (
	*_data,               _size * rows, label_traits<Label>::mpi_type (),
	_data [_local_begin], _size * rows, label_traits<Label>::mpi_type (),
	root);
}

template <typename Label>
void uni_dist_graph<Label>::bcast_row (size_t row)
{
    int root = row / (_size / _comm->Get_size ());
    _comm->Bcast (_data [row], _size, label_traits<Label>::mpi_type (), root);
}

template <typename Label>
void uni_dist_graph<Label>::gather (int root)
{
    if (_comm)
    {
	if (root < 0)
	    root = _default_root;
	
	size_t rows = _local_end - _local_begin;    
	_comm->Gather (
	    _data [_local_begin], _size * rows, label_traits<Label>::mpi_type (),
	    *_data,               _size * rows, label_traits<Label>::mpi_type (),
	    root);
    }
}

} /* namespace pdp */

#endif /* PDP_UNI_DIST_GRAPH_TPP_ */
