/**
 *  Time-stamp:  <2010-12-02 13:41:03 raskolnikov>
 *
 *  @file        graph.tpp
 *  @author      Juan Pedro Bolivar Puente <raskolnikov@es.gnu.org>
 *  @date        Wed Nov 17 21:41:39 2010
 *
 *  Implementación.
 */

#ifndef PDP_GRAPH_TPP
#define PDP_GRAPH_TPP

#include <stdexcept>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <cstdlib>

namespace pdp
{

/**
 * @todo Los operadores << y >> se han implementado siguiendo la
 * implementación del graph.cc original. Sin embargo, una
 * implementación correcta debería satisfacer que los valores escritos
 * y leidos sean equivalentes.
 */
template <typename Label>
std::ostream& operator<< (std::ostream& os, const graph<Label>& g)
{
    for (size_t i = 0; i < g.size (); ++i)
    {
	os << "[" << i << ",*] = ";
	
	for (size_t j = 0; j < g.size (); j++)
	{
	    if (g (i, j) == label_traits<Label>::infinity ()) 
		os << "INF";
	    else  
		os << g (i, j);

	    if (j < g.size () - 1) 
		os << ",";
	    else
		os << std::endl;
	}
    }
    return os;
}

template <typename Label>
std::istream& operator>> (std::istream& is, graph<Label>& g)
{
    size_t graph_size;
    
    if (is >> graph_size)
    {
	g.recreate (graph_size);
 
	size_t i, j, w;
	while ((is >> i) && (is >> j) && (is >> w))
	    g (i, j) = w;
    }
    
    return is;
}

template <typename Label>
graph<Label>::graph (size_t verts)
{
    _allocate (verts);
}

template <typename Label>
graph<Label>::~graph ()
{
    _liberate ();
}

template <typename Label>
void graph<Label>::recreate  (size_t verts)
{
    _liberate ();
    _allocate (verts);
}

template <typename Label>
void graph<Label>::_allocate (size_t s)
{
    _data = 0;
    _size = s;

    if (s)
    {
	_data  = (Label**) std::malloc (sizeof (Label*)*s + sizeof (Label)*s*s);
	*_data = (Label*) (_data + s);
	for (Label** it = _data + 1; it != _data + s; ++it)
	    *it = *(it - 1) + s;
	std::fill (*_data, *_data + s*s, label_traits<Label>::infinity ());
	for (Label* it = *_data; it < *_data + s*s; it += s + 1)
	    *it = 0;
    }
}

template <typename Label>
void graph<Label>::_liberate ()
{
    if (_data)
        std::free (_data);
}

} /* namespace pdp */

#endif /* PDP_GRAPH_TPP */
